/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CvfdhWarpPerspective;
   
DECLARE_NODE(CvfdhWarpPerspective)
/*Node
 *
 * @name CvfdhWarpPerspective
 * @category FDOPENCV_HARK:Transform
 * @description CvfdhWarpPerspective
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name IMAGEOUT
 * @output_type any
 * @output_description Output image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 END*/
   
   
class CvfdhWarpPerspective : public BufferedNode {

  int m_imageInID;
  int m_imageOutID;
      
public:
  CvfdhWarpPerspective(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {

    ObjectRef in = getInput(m_imageInID, count);

    // -------------------------------------------------------------

    if (typeid(*in) == typeid(CvfdImage)) {

    RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
    if(imagePtr->getImage() == NULL){
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
      return;
    }

    IplImage* cv_image;
    cv_image = cvCloneImage(imagePtr->getImage());
    
    IplImage *dst = NULL;
    CvPoint2D32f src_pnt[4], dst_pnt[4];
    CvMat *map_matrix;

    int i;
    double t;
    double a = 29.0, b = 40.0, c = sqrt (a * a + b * b);  /* ߥ顼ѥ᡼ */
    double f = 100;
    int xc = 343, yc = 258;
    int x[4] = { -30, 30, 30, -30 };
    int y[4] = { -50, -50, -50, -50 };
    int z[4] = { 180, 180, 100, 100 };
    int xx[4], yy[4];

    dst = cvCloneImage (cv_image);

    for (i = 0; i < 4; i++) {
      t = -a * a * f / ((b * b + c * c) * z[i] - 2 * b * c * sqrt (x[i] * x[i] + y[i] * y[i] + z[i] * z[i]));
      xx[i] = (int) (x[i] * t + xc + 0.5);
      yy[i] = (int) (y[i] * t + yc + 0.5);
      src_pnt[i] = cvPoint2D32f (xx[i], yy[i]);
    }

    dst_pnt[0] = cvPoint2D32f (0.0, 0.0);
    dst_pnt[1] = cvPoint2D32f (dst->width, 0.0);
    dst_pnt[2] = cvPoint2D32f (dst->width, dst->height);
    dst_pnt[3] = cvPoint2D32f (0.0, dst->height);

    map_matrix = cvCreateMat (3, 3, CV_32FC1);
    cvGetPerspectiveTransform (src_pnt, dst_pnt, map_matrix);
    cvWarpPerspective (cv_image, dst, map_matrix, CV_INTER_LINEAR + CV_WARP_FILL_OUTLIERS, cvScalarAll (0));

    // for (i = 0; i < 4; i++) {
    //   cvLine (cv_image, cvPoint (xx[i], yy[i]), cvPoint (xx[(i + 1) % 4], yy[(i + 1) % 4]), CV_RGB (255, 255, 0), 1, 0, 0);
    // }

    IplImage* img_out;
    img_out = cvCloneImage(dst);

    CvfdImage* image= new CvfdImage(img_out);  
         
    (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

    cvReleaseImage(&dst);
    cvReleaseImage(&cv_image);
    cvReleaseImage(&img_out);

    // -------------------------------------------------------------

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {

      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	IplImage* cv_image;
	cv_image = cvCloneImage(imagePtr.getImage());

	IplImage *dst = NULL;
	CvPoint2D32f src_pnt[4], dst_pnt[4];
	CvMat *map_matrix;

	int i;
	double t;
	double a = 29.0, b = 40.0, c = sqrt (a * a + b * b);  /* ߥ顼ѥ᡼ */
	double f = 100;
	int xc = 343, yc = 258;
	int x[4] = { -30, 30, 30, -30 };
	int y[4] = { -50, -50, -50, -50 };
	int z[4] = { 180, 180, 100, 100 };
	int xx[4], yy[4];
	
	dst = cvCloneImage (cv_image);
	
	for (i = 0; i < 4; i++) {
	  t = -a * a * f / ((b * b + c * c) * z[i] - 2 * b * c * sqrt (x[i] * x[i] + y[i] * y[i] + z[i] * z[i]));
	  xx[i] = (int) (x[i] * t + xc + 0.5);
	  yy[i] = (int) (y[i] * t + yc + 0.5);
	  src_pnt[i] = cvPoint2D32f (xx[i], yy[i]);
	}
	
	dst_pnt[0] = cvPoint2D32f (0.0, 0.0);
	dst_pnt[1] = cvPoint2D32f (dst->width, 0.0);
	dst_pnt[2] = cvPoint2D32f (dst->width, dst->height);
	dst_pnt[3] = cvPoint2D32f (0.0, dst->height);
	
	map_matrix = cvCreateMat (3, 3, CV_32FC1);
	cvGetPerspectiveTransform (src_pnt, dst_pnt, map_matrix);
	cvWarpPerspective (cv_image, dst, map_matrix, CV_INTER_LINEAR + CV_WARP_FILL_OUTLIERS, cvScalarAll (0));
	
	// for (i = 0; i < 4; i++) {
	//   cvLine (cv_image, cvPoint (xx[i], yy[i]), cvPoint (xx[(i + 1) % 4], yy[(i + 1) % 4]), CV_RGB (255, 255, 0), 1, 0, 0);
	// }
	
	IplImage* img_out;
	img_out = cvCloneImage(dst);
	
	CvfdImage* image= new CvfdImage(img_out);  

	(*outputMapPtr)[it->first] = image;

	cvReleaseImage(&dst);
	cvReleaseImage(&cv_image);
	cvReleaseImage(&img_out);

      }
      (*(outputs[m_imageOutID].buffer))[count] = outputMapPtr; 
    }

  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhWarpPerspective)
};
